# SPDX-License-Identifier: BSD-2-Clause

from typing import Literal, Optional

from ..     import (
	ClockDomain, ClockSignal, Elaboratable, Module, ResetSignal, Signal
)

__all__ = (
	'AsyncFFSynchronizer',
	'FFSynchronizer',
	'PulseSynchronizer',
	'ResetSynchronizer',
)


def _check_stages(stages: int):
	if not isinstance(stages, int) or stages < 1:
		raise TypeError(f'Synchronization stage count must be a positive integer, not {stages!r}')
	if stages < 2:
		raise ValueError('Synchronization stage count may not safely be less than 2')


class FFSynchronizer(Elaboratable):
	'''
	Resynchronize a signal to a different clock domain.

	Consists of a chain of flip-flops. Eliminates metastabilities at the output, but provides
	no other guarantee as to the safe domain-crossing of a signal.

	Parameters
	----------
	i : Signal(n), in
		Signal to be resynchronized.
	o : Signal(n), out
		Signal connected to synchronizer output.
	o_domain : str
		Name of output clock domain.
	reset : int
		Reset value of the flip-flops. On FPGAs, even if ``reset_less`` is ``True``,
		the :class:`FFSynchronizer` is still set to this value during initialization.
	reset_less : bool
		If ``True`` (the default), this :class:`FFSynchronizer` is unaffected by ``o_domain``
		reset. See the note below for details.
	stages : int, >=2
		Number of synchronization stages between input and output. The lowest safe number is 2,
		with higher numbers reducing MTBF further, at the cost of increased latency.
	max_input_delay : None or float
		Maximum delay from the input signal's clock to the first synchronization stage, in seconds.
		If specified and the platform does not support it, elaboration will fail.


	.. note::

		:class:`FFSynchronizer` is non-resettable by default. Usually this is the safest option;
		on FPGAs the :class:`FFSynchronizer` will still be initialized to its ``reset`` value when
		the FPGA loads its configuration.

		However, in designs where the value of the :class:`FFSynchronizer` must be valid immediately
		after reset, consider setting ``reset_less`` to ``False`` if any of the following is true:

		- You are targeting an ASIC, or an FPGA that does not allow arbitrary initial flip-flop states;
		- Your design features warm (non-power-on) resets of ``o_domain``, so the one-time
		  initialization at power on is insufficient;
		- Your design features a sequenced reset, and the :class:`FFSynchronizer` must maintain
		  its reset value until ``o_domain`` reset specifically is de-asserted.

		:class:`FFSynchronizer` is reset by the ``o_domain`` reset only.

	Platform overrides
	------------------
	Define the ``get_ff_sync`` platform method to override the implementation of
	:class:`FFSynchronizer`, e.g. to instantiate library cells directly.

	''' # noqa: E101

	def __init__(
		self, i: Signal, o: Signal, *, o_domain: str = 'sync', reset: int = 0,
		reset_less: bool = True, stages: int = 2, max_input_delay: Optional[float] = None
	) -> None:
		_check_stages(stages)

		self.i = i
		self.o = o

		self._reset      = reset
		self._reset_less = reset_less
		self._o_domain   = o_domain
		self._stages     = stages

		self._max_input_delay = max_input_delay

	def elaborate(self, platform) -> Module:
		if hasattr(platform, 'get_ff_sync'):
			return platform.get_ff_sync(self)

		if self._max_input_delay is not None:
			raise NotImplementedError(
				f'Platform \'{type(platform).__name__}\' does not support constraining input delay for FFSynchronizer'
			)

		m = Module()
		flops = [
			Signal(self.i.shape(), name = f'stage{index}', reset = self._reset, reset_less = self._reset_less)
			for index in range(self._stages)
		]
		for i, o in zip((self.i, *flops), flops):
			m.d[self._o_domain] += o.eq(i)
		m.d.comb += self.o.eq(flops[-1])
		return m


class AsyncFFSynchronizer(Elaboratable):
	'''
	Synchronize de-assertion of an asynchronous signal.

	The signal driven by the :class:`AsyncFFSynchronizer` is asserted asynchronously and de-asserted
	synchronously, eliminating metastability during de-assertion.

	This synchronizer is primarily useful for resets and reset-like signals.

	Parameters
	----------
	i : Signal(1), in
		Asynchronous input signal, to be synchronized.
	o : Signal(1), out
		Synchronously released output signal.
	o_domain : str
		Name of clock domain to synchronize to.
	stages : int, >=2
		Number of synchronization stages between input and output. The lowest safe number is 2,
		with higher numbers reducing MTBF further, at the cost of increased de-assertion latency.
	async_edge : str
		The edge of the input signal which causes the output to be set. Must be one of "pos" or "neg".
	max_input_delay : None or float
		Maximum delay from the input signal's clock to the first synchronization stage, in seconds.
		If specified and the platform does not support it, elaboration will fail.

	Platform overrides
	------------------
	Define the ``get_async_ff_sync`` platform method to override the implementation of
	:class:`AsyncFFSynchronizer`, e.g. to instantiate library cells directly.

	'''

	def __init__(
		self, i: Signal, o: Signal, *, o_domain: str = 'sync', stages: int = 2,
		async_edge: Literal['pos', 'neg'] = 'pos', max_input_delay: Optional[float] = None
	) -> None:
		_check_stages(stages)

		if len(i) != 1:
			raise ValueError(f'AsyncFFSynchronizer input width must be 1, not {len(i)}')
		if len(o) != 1:
			raise ValueError(f'AsyncFFSynchronizer output width must be 1, not {len(o)}')

		if async_edge not in ('pos', 'neg'):
			raise ValueError(f'AsyncFFSynchronizer async edge must be one of \'pos\' or \'neg\', not {async_edge!r}')

		self.i = i
		self.o = o

		self._o_domain = o_domain
		self._stages = stages

		self._edge = async_edge

		self._max_input_delay = max_input_delay

	def elaborate(self, platform) -> Module:
		if hasattr(platform, 'get_async_ff_sync'):
			return platform.get_async_ff_sync(self)

		if self._max_input_delay is not None:
			raise NotImplementedError(
				f'Platform \'{type(platform).__name__}\' does not support constraining input delay for AsyncFFSynchronizer'
			)

		m = Module()
		m.domains += ClockDomain('async_ff', async_reset = True, local = True)
		flops = [
			Signal(1, name = f'stage{index}', reset = 1)
			for index in range(self._stages)
		]
		for i, o in zip((0, *flops), flops):
			m.d.async_ff += o.eq(i)

		if self._edge == 'pos':
			m.d.comb += ResetSignal('async_ff').eq(self.i)
		else:
			m.d.comb += ResetSignal('async_ff').eq(~self.i)

		m.d.comb += [
			ClockSignal('async_ff').eq(ClockSignal(self._o_domain)),
			self.o.eq(flops[-1])
		]

		return m


class ResetSynchronizer(Elaboratable):
	'''
	Synchronize de-assertion of a clock domain reset.

	The reset of the clock domain driven by the :class:`ResetSynchronizer` is asserted
	asynchronously and de-asserted synchronously, eliminating metastability during de-assertion.

	The driven clock domain could use a reset that is asserted either synchronously or
	asynchronously; a reset is always de-asserted synchronously. A domain with an asynchronously
	asserted reset is useful if the clock of the domain may be gated, yet the domain still
	needs to be reset promptly; otherwise, synchronously asserted reset (the default) should
	be used.

	Parameters
	----------
	arst : Signal(1), in
		Asynchronous reset signal, to be synchronized.
	domain : str
		Name of clock domain to reset.
	stages : int, >=2
		Number of synchronization stages between input and output. The lowest safe number is 2,
		with higher numbers reducing MTBF further, at the cost of increased de-assertion latency.
	max_input_delay : None or float
		Maximum delay from the input signal's clock to the first synchronization stage, in seconds.
		If specified and the platform does not support it, elaboration will fail.

	Platform overrides
	------------------
	Define the ``get_reset_sync`` platform method to override the implementation of
	:class:`ResetSynchronizer`, e.g. to instantiate library cells directly.

	'''

	def __init__(
		self, arst: Signal, *, domain: str = 'sync', stages: int = 2,
		max_input_delay: Optional[float] = None
	) -> None:
		_check_stages(stages)

		self.arst = arst

		self._domain = domain
		self._stages = stages

		self._max_input_delay = max_input_delay

	def elaborate(self, platform) -> AsyncFFSynchronizer:
		return AsyncFFSynchronizer(
			self.arst,
			ResetSignal(self._domain),
			o_domain = self._domain,
			stages = self._stages,
			max_input_delay = self._max_input_delay
		)


class PulseSynchronizer(Elaboratable):
	'''
	A one-clock pulse on the input produces a one-clock pulse on the output.

	If the output clock is faster than the input clock, then the input may be safely asserted at
	100% duty cycle. Otherwise, if the clock ratio is ``n``:1, the input may be asserted at most
	once in every ``n`` input clocks, else pulses may be dropped. Other than this there is
	no constraint on the ratio of input and output clock frequency.

	Parameters
	----------
	i_domain : str
		Name of input clock domain.
	o_domain : str
		Name of output clock domain.
	stages : int, >=2
		Number of synchronization stages between input and output. The lowest safe number is 2,
		with higher numbers reducing MTBF further, at the cost of increased de-assertion latency.

	'''

	def __init__(self, i_domain: str, o_domain: str, *, stages: int = 2) -> None:
		_check_stages(stages)

		self.i = Signal()
		self.o = Signal()

		self._i_domain = i_domain
		self._o_domain = o_domain
		self._stages = stages

	def elaborate(self, platform) -> Module:
		m = Module()

		i_toggle = Signal()
		o_toggle = Signal()
		r_toggle = Signal()
		ff_sync = m.submodules.ff_sync = FFSynchronizer( # noqa: F841
			i_toggle, o_toggle, o_domain = self._o_domain, stages = self._stages
		)

		m.d[self._i_domain] += i_toggle.eq(i_toggle ^ self.i)
		m.d[self._o_domain] += r_toggle.eq(o_toggle)
		m.d.comb += self.o.eq(o_toggle ^ r_toggle)

		return m



# SPDX-License-Identifier: BSD-2-Clause

from .gowin      import GowinPlatform
from .intel      import IntelPlatform
from .lattice    import (
	LatticeECP5Platform, LatticeICE40Platform, LatticeMachXO2Platform, LatticeMachXO3LPlatform
)
from .quicklogic import QuicklogicPlatform
from .xilinx     import XilinxPlatform

__all__ = (
	'GowinPlatform',

	'IntelPlatform',

	'LatticeECP5Platform',
	'LatticeICE40Platform',
	'LatticeMachXO2Platform',
	'LatticeMachXO3LPlatform',

	'QuicklogicPlatform',

	'XilinxPlatform',
)



# SPDX-License-Identifier: BSD-2-Clause

from setuptools.command.build_ext import new_compiler
from os                           import chdir, getcwd
from sysconfig                    import get_config_var
from tempfile                     import TemporaryDirectory

__all__ = (
	'build_cxx',
)

# TODO: Refactor this so you don't need to pass the sources along with the filename in if not needed
def build_cxx(
	cxx_sources: dict[str, str], output_name: str, include_dirs: list[str], macros: list[str]
)  -> tuple[TemporaryDirectory, str]:

	build_dir = TemporaryDirectory(prefix = 'torii_cxx_')

	cwd = getcwd()
	try:
		# Unfortunately, `ccompiler.compile` assumes the paths are relative, and interprets
		# the directory name of the source path specially. That makes it necessary to build in
		# the output directory directly.
		chdir(build_dir.name)

		cc_driver = new_compiler()

		cc_driver.output_dir = '.'

		cc          = get_config_var('CC')
		cxx         = get_config_var('CXX')
		cflags      = get_config_var('CCSHARED')
		ld_flags    = get_config_var('LDSHARED')
		ld_cxxflags = get_config_var('LDCXXSHARED')
		if ld_cxxflags is None:
			# PyPy doesn't have LDCXXSHARED. Glue it together from CXX and LDSHARED and hope that
			# the result actually works; not many good options here.
			ld_cxxflags = ' '.join([cxx.split()[0], *ld_flags.split()[1:]])
		cc_driver.set_executables(
			compiler     = f'{cc} {cflags}',
			compiler_so  = f'{cc} {cflags}',
			compiler_cxx = f'{cxx} {cflags}',
			linker_so    = ld_cxxflags,
		)

		# Sometimes CCompiler is modified to have additional executable entries for compiling and
		# linking CXX shared objects (e.g. on Gentoo). These executables have to be set then.
		try:
			cc_driver.set_executables(
				compiler_so_cxx = f'{cxx} {cflags}',
				linker_so_cxx   = ld_cxxflags,
			)
		except: # noqa: E722
			pass

		for include_dir in include_dirs:
			cc_driver.add_include_dir(include_dir)
		for macro in macros:
			cc_driver.define_macro(macro)
		for cxx_filename, cxx_source in cxx_sources.items():
			with open(cxx_filename, 'w') as f:
				f.write(cxx_source)

		cxx_filenames = list(cxx_sources.keys())
		obj_filenames = cc_driver.object_filenames(cxx_filenames)
		so_filename   = cc_driver.shared_object_filename(output_name)

		cc_driver.compile(cxx_filenames)
		cc_driver.link_shared_object(obj_filenames, output_filename = so_filename, target_lang = 'c++')

		return (build_dir, so_filename)

	finally:
		chdir(cwd)



# SPDX-License-Identifier: BSD-2-Clause

from ... import Elaboratable, Module, Signal

__all__ = (
	'Decoder',
	'Encoder',
)

class Encoder(Elaboratable):
	'''
	Encode one-hot to binary.

	If one bit in ``i`` is asserted, ``n`` is low and ``o`` indicates the asserted bit.
	Otherwise, ``n`` is high and ``o`` is ``0``.

	Parameters
	----------
	width : int
		Bit width of the input

	Attributes
	----------
	i : Signal(width), in
		One-hot input.
	o : Signal(range(width)), out
		Encoded natural binary.
	n : Signal, out
		Invalid: either none or multiple input bits are asserted.

	'''

	def __init__(self, width: int) -> None:
		self.width = width

		self.i = Signal(width)
		self.o = Signal(range(width))
		self.n = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		with m.Switch(self.i):
			for j in range(self.width):
				with m.Case(1 << j):
					m.d.comb += self.o.eq(j)
			with m.Case():
				m.d.comb += self.n.eq(1)
		return m

class Decoder(Elaboratable):
	'''
	Decode binary to one-hot.

	If ``n`` is low, only the ``i``-th bit in ``o`` is asserted.
	If ``n`` is high, ``o`` is ``0``.

	Parameters
	----------
	width : int
		Bit width of the output.

	Attributes
	----------
	i : Signal(range(width)), in
		Input binary.
	o : Signal(width), out
		Decoded one-hot.
	n : Signal, in
		Invalid, no output bits are to be asserted.

	'''

	def __init__(self, width: int) -> None:
		self.width = width

		self.i = Signal(range(width))
		self.n = Signal()
		self.o = Signal(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		with m.Switch(self.i):
			for j in range(len(self.o)):
				with m.Case(j):
					m.d.comb += self.o.eq(1 << j)
		with m.If(self.n):
			m.d.comb += self.o.eq(0)
		return m



# SPDX-License-Identifier: BSD-2-Clause

from enum     import Enum
from typing   import Optional

from ....     import Elaboratable, Module, Mux, Record, Signal
from ..memory import MemoryMap

__all__ = (
	'Decoder',
	'Element',
	'Interface',
	'Multiplexer',
)

class Element(Record):
	class Access(Enum):
		'''
		Register access mode.

		Coarse access mode for the entire register. Individual fields can have more restrictive
		access mode, e.g. R/O fields can be a part of an R/W register.

		'''

		R  = 'r'
		W  = 'w'
		RW = 'rw'

		def readable(self) -> bool:
			return self == self.R or self == self.RW

		def writable(self) -> bool:
			return self == self.W or self == self.RW

	'''
	Peripheral-side CSR interface.

	A low-level interface to a single atomically readable and writable register in a peripheral.
	This interface supports any register width and semantics, provided that both reads and writes
	always succeed and complete in one cycle.

	Parameters
	----------
	width : int
		Width of the register.
	access : :class:`Access`
		Register access mode.
	name : str
		Name of the underlying record.

	Attributes
	----------
	r_data : Signal(width)
		Read data. Must be always valid, and is sampled when ``r_stb`` is asserted.
	r_stb : Signal()
		Read strobe. Registers with read side effects should perform the read side effect when this
		strobe is asserted.
	w_data : Signal(width)
		Write data. Valid only when ``w_stb`` is asserted.
	w_stb : Signal()
		Write strobe. Registers should update their value or perform the write side effect when
		this strobe is asserted.

	'''

	def __init__(
		self, width: int, access: Access, *, name: Optional[str] = None, src_loc_at: int = 0
	) -> None:
		if not isinstance(width, int) or width < 0:
			raise ValueError(f'Width must be a non-negative integer, not {width!r}')
		if not isinstance(access, Element.Access) and access not in ('r', 'w', 'rw'):
			raise ValueError(f'Access mode must be one of \'r\', \'w\', or \'rw\', not {access!r}')
		self.width  = width
		self.access = Element.Access(access)

		layout = []
		if self.access.readable():
			layout += [
				('r_data', width),
				('r_stb',  1),
			]
		if self.access.writable():
			layout += [
				('w_data', width),
				('w_stb',  1),
			]
		super().__init__(layout, name = name, src_loc_at = 1 + src_loc_at)


class Interface(Record):
	'''
	CPU-side CSR interface.

	A low-level interface to a set of atomically readable and writable peripheral CSR registers.

	Operation
	---------

	CSR registers mapped to the CSR bus are split into chunks according to the bus data width.
	Each chunk is assigned a consecutive address on the bus. This allows accessing CSRs of any
	size using any datapath width.

	When the first chunk of a register is read, the value of a register is captured, and reads
	from subsequent chunks of the same register return the captured values. When any chunk except
	the last chunk of a register is written, the written value is captured; a write to the last
	chunk writes the captured value to the register. This allows atomically accessing CSRs larger
	than datapath width.

	Parameters
	----------
	addr_width : int
		Address width. At most ``(2 ** addr_width) * data_width`` register bits will be available.
	data_width : int
		Data width. Registers are accessed in ``data_width`` sized chunks.
	name : str
		Name of the underlying record.

	Attributes
	----------
	memory_map : MemoryMap
		Map of the bus.
	addr : Signal(addr_width)
		Address for reads and writes.
	r_data : Signal(data_width)
		Read data. Valid on the next cycle after ``r_stb`` is asserted. Otherwise, zero. (Keeping
		read data of an unused interface at zero simplifies multiplexers.)
	r_stb : Signal()
		Read strobe. If ``addr`` points to the first chunk of a register, captures register value
		and causes read side effects to be performed (if any). If ``addr`` points to any chunk
		of a register, latches the captured value to ``r_data``. Otherwise, latches zero
		to ``r_data``.
	w_data : Signal(data_width)
		Write data. Must be valid when ``w_stb`` is asserted.
	w_stb : Signal()
		Write strobe. If ``addr`` points to the last chunk of a register, writes captured value
		to the register and causes write side effects to be performed (if any). If ``addr`` points
		to any chunk of a register, latches ``w_data`` to the captured value. Otherwise, does
		nothing.

	'''

	def __init__(self, *, addr_width: int, data_width: int, name: Optional[str] = None) -> None:
		if not isinstance(addr_width, int) or addr_width <= 0:
			raise ValueError(f'Address width must be a positive integer, not {addr_width!r}')
		if not isinstance(data_width, int) or data_width <= 0:
			raise ValueError(f'Data width must be a positive integer, not {data_width!r}')
		self.addr_width = addr_width
		self.data_width = data_width
		self._map       = None

		super().__init__([
			('addr',    addr_width),
			('r_data',  data_width),
			('r_stb',   1),
			('w_data',  data_width),
			('w_stb',   1),
		], name = name, src_loc_at = 1)

	@property
	def memory_map(self) -> MemoryMap:
		if self._map is None:
			raise NotImplementedError(f'Bus interface {self!r} does not have a memory map')
		return self._map

	@memory_map.setter
	def memory_map(self, memory_map: MemoryMap) -> None:
		if not isinstance(memory_map, MemoryMap):
			raise TypeError(f'Memory map must be an instance of MemoryMap, not {memory_map!r}')
		if memory_map.addr_width != self.addr_width:
			raise ValueError(
				f'Memory map has address width {memory_map.addr_width}, '
				f'which is not the same as bus interface address width {self.addr_width}'
			)
		if memory_map.data_width != self.data_width:
			raise ValueError(
				f'Memory map has data width {memory_map.data_width}, '
				f'which is not the same as bus interface data width {self.data_width}'
			)
		memory_map.freeze()
		self._map = memory_map


class Multiplexer(Elaboratable):
	'''
	CSR register multiplexer.

	An address-based multiplexer for CSR registers implementing atomic updates.

	Latency
	-------

	Writes are registered, and are performed 1 cycle after ``w_stb`` is asserted.

	Alignment
	---------

	Because the CSR bus conserves logic and routing resources, it is common to e.g. access
	a CSR bus with an *n*-bit data path from a CPU with a *k*-bit datapath (*k>n*) in cases
	where CSR access latency is less important than resource usage. In this case, two strategies
	are possible for connecting the CSR bus to the CPU:

		* The CPU could access the CSR bus directly (with no intervening logic other than simple
		  translation of control signals). In this case, the register alignment should be set
		  to 1 (i.e. `alignment` should be set to 0), and each *w*-bit register would occupy
		  *ceil(w/n)* addresses from the CPU perspective, requiring the same amount of memory
		  instructions to access.
		* The CPU could also access the CSR bus through a width down-converter, which would issue
		  *k/n* CSR accesses for each CPU access. In this case, the register alignment should be
		  set to *k/n*, and each *w*-bit register would occupy *ceil(w/k)* addresses from the CPU
		  perspective, requiring the same amount of memory instructions to access.

	If the register alignment (i.e. `2 ** alignment`) is greater than 1, it affects which CSR bus
	write is considered a write to the last register chunk. For example, if a 24-bit register is
	used with a 8-bit CSR bus and a CPU with a 32-bit datapath, a write to this register requires
	4 CSR bus writes to complete and the 4th write is the one that actually writes the value to
	the register. This allows determining write latency solely from the amount of addresses the
	register occupies in the CPU address space, and the width of the CSR bus.

	Parameters
	----------
	addr_width : int
		Address width. See :class:`Interface`.
	data_width : int
		Data width. See :class:`Interface`.
	alignment : log2 of int
		Register alignment. See :class:`..memory.MemoryMap`.
	name : str
		Window name. Optional.

	Attributes
	----------
	bus : :class:`Interface`
		CSR bus providing access to registers.

	''' # noqa: E101

	def __init__(
		self, *, addr_width: int, data_width: int, alignment: int = 0, name: Optional[str] = None
	) -> None:
		self._map = MemoryMap(
			addr_width = addr_width, data_width = data_width,
			alignment = alignment, name = name
		)
		self._bus = None

	@property
	def bus(self) -> Interface:
		if self._bus is None:
			self._map.freeze()
			self._bus = Interface(
				addr_width = self._map.addr_width,
				data_width = self._map.data_width,
				name = 'csr'
			)
			self._bus.memory_map = self._map
		return self._bus

	def align_to(self, alignment: int) -> int:
		'''
		Align the implicit address of the next register.

		See :meth:`MemoryMap.align_to` for details.

		'''

		return self._map.align_to(alignment)

	def add(
		self, element: Element, *, addr: Optional[int] = None, alignment: Optional[int] = None,
		extend: bool = False
	) -> tuple[int, int]:
		'''
		Add a register.

		See :meth:`MemoryMap.add_resource` for details.

		'''

		if not isinstance(element, Element):
			raise TypeError(f'Element must be an instance of csr.Element, not {element!r}')

		size = (element.width + self._map.data_width - 1) // self._map.data_width
		return self._map.add_resource(
			element, size = size, addr = addr, alignment = alignment,
			extend = extend, name = element.name
		)

	def elaborate(self, platform) -> Module:
		m = Module()

		# Instead of a straightforward multiplexer for reads, use a per-element address comparator,
		# AND the shadow register chunk with the comparator output, and OR all of those together.
		# If the toolchain doesn't already synthesize multiplexer trees this way, this trick can
		# save a significant amount of logic, since e.g. one 4-LUT can pack one 2-MUX, but two
		# 2-AND or 2-OR gates.
		r_data_fanin = 0

		for elem, _, (elem_start, elem_end) in self._map.resources():
			shadow = Signal(elem.width, name = f'{elem.name}__shadow')
			if elem.access.readable():
				shadow_en = Signal(elem_end - elem_start, name = f'{elem.name}__shadow_en')
				m.d.sync += shadow_en.eq(0)
			if elem.access.writable():
				m.d.comb += elem.w_data.eq(shadow)
				m.d.sync += elem.w_stb.eq(0)

			# Enumerate every address used by the register explicitly, rather than using
			# arithmetic comparisons, since some toolchains (e.g. Yosys) are too eager to infer
			# carry chains for comparisons, even with a constant. (Register sizes don't have
			# to be powers of 2.)
			with m.Switch(self.bus.addr):
				for chunk_offset, chunk_addr in enumerate(range(elem_start, elem_end)):
					shadow_slice = shadow.word_select(chunk_offset, self.bus.data_width)

					with m.Case(chunk_addr):
						if elem.access.readable():
							r_data_fanin |= Mux(shadow_en[chunk_offset], shadow_slice, 0)
							if chunk_addr == elem_start:
								m.d.comb += elem.r_stb.eq(self.bus.r_stb)
								with m.If(self.bus.r_stb):
									m.d.sync += shadow.eq(elem.r_data)
							# Delay by 1 cycle, allowing reads to be pipelined.
							m.d.sync += shadow_en.eq(self.bus.r_stb << chunk_offset)

						if elem.access.writable():
							if chunk_addr == elem_end - 1:
								# Delay by 1 cycle, avoiding combinatorial paths through
								# the CSR bus and into CSR registers.
								m.d.sync += elem.w_stb.eq(self.bus.w_stb)
							with m.If(self.bus.w_stb):
								m.d.sync += shadow_slice.eq(self.bus.w_data)

		m.d.comb += self.bus.r_data.eq(r_data_fanin)

		return m


class Decoder(Elaboratable):
	'''
	CSR bus decoder.

	An address decoder for subordinate CSR buses.

	Usage
	-----

	Although there is no functional difference between adding a set of registers directly to
	a :class:`Multiplexer` and adding a set of registers to multiple :class:`Multiplexer`s that are
	aggregated with a :class:`Decoder`, hierarchical CSR buses are useful for organizing
	a hierarchical design. If many peripherals are directly served by a single
	:class:`Multiplexer`, a very large amount of ports will connect the peripheral registers with
	the decoder, and the cost of decoding logic would not be attributed to specific peripherals.
	With a decoder, only five signals per peripheral will be used, and the logic could be kept
	together with the peripheral.

	Parameters
	----------
	addr_width : int
		Address width. See :class:`Interface`.
	data_width : int
		Data width. See :class:`Interface`.
	alignment : log2 of int
		Window alignment. See :class:`..memory.MemoryMap`.
	name : str
		Window name. Optional.

	Attributes
	----------
	bus : :class:`Interface`
		CSR bus providing access to subordinate buses.

	'''

	def __init__(
		self, *, addr_width: int, data_width: int, alignment: int = 0, name: Optional[str] = None
	) -> None:
		self._map  = MemoryMap(
			addr_width = addr_width, data_width = data_width,
			alignment = alignment, name = name,
		)
		self._bus  = None
		self._subs = dict()

	@property
	def bus(self) -> Interface:
		if self._bus is None:
			self._map.freeze()
			self._bus = Interface(
				addr_width = self._map.addr_width,
				data_width = self._map.data_width,
				name = 'csr'
			)
			self._bus.memory_map = self._map
		return self._bus

	def align_to(self, alignment: int) -> int:
		'''
		Align the implicit address of the next window.

		See :meth:`MemoryMap.align_to` for details.

		'''

		return self._map.align_to(alignment)

	def add(
		self, sub_bus: Interface, *, addr: Optional[int] = None, extend: bool = False
	) -> tuple[int, int, int]:
		'''
		Add a window to a subordinate bus.

		See :meth:`MemoryMap.add_resource` for details.

		'''

		if not isinstance(sub_bus, Interface):
			raise TypeError(f'Subordinate bus must be an instance of csr.Interface, not {sub_bus!r}')
		if sub_bus.data_width != self._map.data_width:
			raise ValueError(
				f'Subordinate bus has data width {sub_bus.data_width}, '
				f'which is not the same as decoder data width {self._map.data_width}'
			)
		self._subs[sub_bus.memory_map] = sub_bus
		return self._map.add_window(sub_bus.memory_map, addr = addr, extend = extend)

	def elaborate(self, platform) -> Module:
		m = Module()

		# See Multiplexer.elaborate above.
		r_data_fanin = 0

		with m.Switch(self.bus.addr):
			for sub_map, (sub_pat, sub_ratio) in self._map.window_patterns():
				assert sub_ratio == 1

				sub_bus = self._subs[sub_map]
				m.d.comb += sub_bus.addr.eq(self.bus.addr[:sub_bus.addr_width])

				# The CSR bus interface is defined to output zero when idle, allowing us to avoid
				# adding a multiplexer here.
				r_data_fanin |= sub_bus.r_data
				m.d.comb += sub_bus.w_data.eq(self.bus.w_data)

				with m.Case(sub_pat):
					m.d.comb += sub_bus.r_stb.eq(self.bus.r_stb)
					m.d.comb += sub_bus.w_stb.eq(self.bus.w_stb)

		m.d.comb += self.bus.r_data.eq(r_data_fanin)

		return m



# SPDX-License-Identifier: BSD-2-Clause

from inspect   import getfile, getlineno, iscoroutine, isgenerator

from ..hdl     import ClockDomain, Const, Value
from ..hdl.ast import SignalSet, Statement
from ._base    import BaseProcess
from ._pyrtl   import _RHSValueCompiler, _StatementCompiler, _ValueCompiler
from .core     import Active, Delay, Passive, Settle, Tick

__all__ = (
    'PyCoroProcess',
)


class PyCoroProcess(BaseProcess):
    def __init__(self, state, domains, constructor, *, default_cmd = None):
        self.state = state
        self.domains = domains
        self.constructor = constructor
        self.default_cmd = default_cmd

        self.reset()

    def reset(self):
        self.runnable = True
        self.passive = False

        self.coroutine = self.constructor()
        self.exec_locals = {
            'slots': self.state.slots,
            'result': None,
            **_ValueCompiler.helpers
        }
        self.waits_on = SignalSet()

    def src_loc(self):
        coroutine = self.coroutine
        if coroutine is None:
            return None
        while coroutine.gi_yieldfrom is not None and isgenerator(coroutine.gi_yieldfrom):
            coroutine = coroutine.gi_yieldfrom
        if isgenerator(coroutine):
            frame = coroutine.gi_frame
        if iscoroutine(coroutine):
            frame = coroutine.cr_frame
        return f'{getfile(frame)}:{getlineno(frame)}'

    def add_trigger(self, signal, trigger = None):
        self.state.add_trigger(self, signal, trigger = trigger)
        self.waits_on.add(signal)

    def clear_triggers(self):
        for signal in self.waits_on:
            self.state.remove_trigger(self, signal)
        self.waits_on.clear()

    def run(self):
        if self.coroutine is None:
            return

        self.clear_triggers()

        response = None
        while True:
            try:
                command = self.coroutine.send(response)
                if command is None:
                    command = self.default_cmd
                response = None

                if isinstance(command, Value):
                    exec(_RHSValueCompiler.compile(self.state, command, mode = 'curr'),
                        self.exec_locals)
                    response = Const.normalize(self.exec_locals['result'], command.shape())

                elif isinstance(command, Statement):
                    exec(_StatementCompiler.compile(self.state, command),
                        self.exec_locals)

                elif type(command) is Tick:
                    domain = command.domain
                    if isinstance(domain, ClockDomain):
                        pass
                    elif domain in self.domains:
                        domain = self.domains[domain]
                    else:
                        raise NameError(f'Received command {command!r} that refers to a nonexistent '
                                        f'domain {command.domain!r} from process {self.src_loc()!r}')
                    self.add_trigger(domain.clk, trigger = 1 if domain.clk_edge == 'pos' else 0)
                    if domain.rst is not None and domain.async_reset:
                        self.add_trigger(domain.rst, trigger = 1)
                    return

                elif type(command) is Settle:
                    self.state.wait_interval(self, None)
                    return

                elif type(command) is Delay:
                    # Internal timeline is in 1ps integeral units, intervals are public API and in floating point
                    interval = int(command.interval * 1e12) if command.interval is not None else None
                    self.state.wait_interval(self, interval)
                    return

                elif type(command) is Passive:
                    self.passive = True

                elif type(command) is Active:
                    self.passive = False

                elif command is None: # only possible if self.default_cmd is None
                    raise TypeError(f'Received default command from process {self.src_loc()!r} that was added '
                                    'with add_process(); did you mean to add this process with '
                                    'add_sync_process() instead?')

                else:
                    raise TypeError(f'Received unsupported command {command!r} from process {self.src_loc()!r}')

            except StopIteration:
                self.passive = True
                self.coroutine = None
                return

            except Exception as exn:
                self.coroutine.throw(exn)



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, EnableInserter
from torii.sim  import Simulator
from torii.back import verilog


class Counter(Elaboratable):
	def __init__(self, width: int):
		self.v = Signal(width, reset = 2**width-1)
		self.o = Signal()
		self.en = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		m.d.sync += self.v.eq(self.v + 1)
		m.d.comb += self.o.eq(self.v[-1])
		return EnableInserter(self.en)(m)


ctr = Counter(width = 16)

print(verilog.convert(ctr, ports = [ctr.o, ctr.en]))

sim = Simulator(ctr)
sim.add_clock(1e-6)
def ce_proc():
	yield
	yield
	yield
	yield ctr.en.eq(1)
	yield
	yield
	yield
	yield ctr.en.eq(0)
	yield
	yield
	yield
	yield ctr.en.eq(1)

sim.add_sync_process(ce_proc)

with sim.write_vcd('ctrl.vcd', 'ctrl.gtkw', traces = [ctr.en, ctr.v, ctr.o]):
	sim.run_until(100e-6, run_passive = True)



# SPDX-License-Identifier: BSD-2-Clause

from typing   import Optional, Union

from ...build import Attrs, Pins, Resource, Subsignal

__all__ = (
	'Display7SegResource',
	'VGADACResource',
	'VGAResource',
)


def Display7SegResource(
	*args,
	a: str, b: str, c: str, d: str, e: str, f: str, g: str, dp: Optional[str] = None,
	invert: bool = False, conn: Optional[Union[tuple[str, int], str]] = None,
	attrs: Optional[Attrs] = None
) -> Resource:

	ios = []

	ios.append(Subsignal('a', Pins(a, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('b', Pins(b, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('c', Pins(c, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('d', Pins(d, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('e', Pins(e, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('f', Pins(f, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('g', Pins(g, dir = 'o', invert = invert, conn = conn, assert_width = 1)))

	if dp is not None:
		ios.append(Subsignal(
			'dp', Pins(dp, dir = 'o', invert = invert, conn = conn, assert_width = 1)
		))

	if attrs is not None:
		ios.append(attrs)
	return Resource.family(*args, default_name = 'display_7seg', ios = ios)


def VGAResource(
	*args,
	r: str, g: str, b: str, vs: str, hs: str,
	invert_sync: bool = False, conn: Optional[Union[tuple[str, int], str]] = None,
	attrs: Optional[Attrs] = None
) -> Resource:

	ios = []

	ios.append(Subsignal('r', Pins(r, dir = 'o', conn = conn)))
	ios.append(Subsignal('g', Pins(g, dir = 'o', conn = conn)))
	ios.append(Subsignal('b', Pins(b, dir = 'o', conn = conn)))
	ios.append(Subsignal(
		'hs', Pins(hs, dir = 'o', invert = invert_sync, conn = conn, assert_width = 1)
	))
	ios.append(Subsignal(
		'vs', Pins(vs, dir = 'o', invert = invert_sync, conn = conn, assert_width = 1)
	))

	if attrs is not None:
		ios.append(attrs)

	return Resource.family(*args, default_name = 'vga', ios = ios)


def VGADACResource(
	*args,
	clk: str, r: str, g: str, b: str, vs: str, hs: str, extras: list[Subsignal] = [],
	invert_sync: bool = False, conn: Optional[Union[tuple[str, int], str]] = None,
	attrs: Optional[Attrs] = None
) -> Resource:

	ios = []

	ios.append(Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('r', Pins(r, dir = 'o', conn = conn)))
	ios.append(Subsignal('g', Pins(g, dir = 'o', conn = conn)))
	ios.append(Subsignal('b', Pins(b, dir = 'o', conn = conn)))
	ios.append(Subsignal(
		'hs', Pins(hs, dir = 'o', invert = invert_sync, conn = conn, assert_width = 1)
	))
	ios.append(Subsignal(
		'vs', Pins(vs, dir = 'o', invert = invert_sync, conn = conn, assert_width = 1)
	))
	for extra in extras:
		ios.append(extra)

	if attrs is not None:
		ios.append(attrs)

	return Resource.family(*args, default_name = 'vgadac', ios = ios)



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal
from torii.back import verilog


class ParMux(Elaboratable):
	def __init__(self, width):
		self.s = Signal(3)
		self.a = Signal(width)
		self.b = Signal(width)
		self.c = Signal(width)
		self.o = Signal(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		with m.Switch(self.s):
			with m.Case('--1'):
				m.d.comb += self.o.eq(self.a)
			with m.Case('-1-'):
				m.d.comb += self.o.eq(self.b)
			with m.Case('1--'):
				m.d.comb += self.o.eq(self.c)
			with m.Case():
				m.d.comb += self.o.eq(0)
		return m


if __name__ == '__main__':
	pmux = ParMux(width = 16)

	print(verilog.convert(pmux, ports = [pmux.s, pmux.a, pmux.b, pmux.c, pmux.o]))



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl                 import *
from torii.asserts             import *
from torii.sim                 import *
from torii.lib.coding.priority import Encoder

from ...utils                  import ToriiTestSuiteCase


class PriorityEncoderTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		enc = Encoder(4)

		def process():
			self.assertEqual((yield enc.n), 1)
			self.assertEqual((yield enc.o), 0)

			yield enc.i.eq(0b0001)
			yield Settle()
			self.assertEqual((yield enc.n), 0)
			self.assertEqual((yield enc.o), 0)

			yield enc.i.eq(0b0100)
			yield Settle()
			self.assertEqual((yield enc.n), 0)
			self.assertEqual((yield enc.o), 2)

			yield enc.i.eq(0b0110)
			yield Settle()
			self.assertEqual((yield enc.n), 0)
			self.assertEqual((yield enc.o), 1)

		sim = Simulator(enc)
		sim.add_process(process)
		sim.run()



# SPDX-License-Identifier: BSD-2-Clause

from abc        import abstractmethod
from typing     import Literal, Optional, Union

from ...build   import Attrs, TemplatedPlatform
from ...hdl     import (
	ClockDomain, ClockSignal, Const, Instance, Module, Record, ResetSignal, Signal
)
from ...lib.cdc import ResetSynchronizer
from ...lib.io  import Pin

__all__ = (
	'XilinxPlatform',
)


class XilinxPlatform(TemplatedPlatform):
	'''
	.. rubric:: Vivado toolchain

	Required tools:
		* ``vivado``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_VIVADO``, if present.

	Available overrides:
		* ``script_after_read``: inserts commands after ``read_xdc`` in Tcl script.
		* ``synth_design_opts``: sets options for ``synth_design``.
		* ``script_after_synth``: inserts commands after ``synth_design`` in Tcl script.
		* ``script_after_place``: inserts commands after ``place_design`` in Tcl script.
		* ``script_after_route``: inserts commands after ``route_design`` in Tcl script.
		* ``script_before_bitstream``: inserts commands before ``write_bitstream`` in Tcl script.
		* ``script_after_bitstream``: inserts commands after ``write_bitstream`` in Tcl script.
		* ``add_constraints``: inserts commands in XDC file.
		* ``vivado_opts``: adds extra options for ``vivado``.

	Build products:
		* ``{{name}}.log``: Vivado log.
		* ``{{name}}_timing_synth.rpt``: Vivado report.
		* ``{{name}}_utilization_hierarchical_synth.rpt``: Vivado report.
		* ``{{name}}_utilization_synth.rpt``: Vivado report.
		* ``{{name}}_utilization_hierarchical_place.rpt``: Vivado report.
		* ``{{name}}_utilization_place.rpt``: Vivado report.
		* ``{{name}}_io.rpt``: Vivado report.
		* ``{{name}}_control_sets.rpt``: Vivado report.
		* ``{{name}}_clock_utilization.rpt``:  Vivado report.
		* ``{{name}}_route_status.rpt``: Vivado report.
		* ``{{name}}_drc.rpt``: Vivado report.
		* ``{{name}}_methodology.rpt``: Vivado report.
		* ``{{name}}_timing.rpt``: Vivado report.
		* ``{{name}}_power.rpt``: Vivado report.
		* ``{{name}}_route.dcp``: Vivado design checkpoint.
		* ``{{name}}.bit``: binary bitstream with metadata.
		* ``{{name}}.bin``: binary bitstream.

	.. rubric:: ISE toolchain

	Required tools:
		* ``xst``
		* ``ngdbuild``
		* ``map``
		* ``par``
		* ``bitgen``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_ISE``, if present.

	Available overrides:
		* ``script_after_run``: inserts commands after ``run`` in XST script.
		* ``add_constraints``: inserts commands in UCF file.
		* ``xst_opts``: adds extra options for ``xst``.
		* ``ngdbuild_opts``: adds extra options for ``ngdbuild``.
		* ``map_opts``: adds extra options for ``map``.
		* ``par_opts``: adds extra options for ``par``.
		* ``bitgen_opts``: adds extra and overrides default options for ``bitgen``;
		  default options: ``-g Compress``.

	Build products:
		* ``{{name}}.srp``: synthesis report.
		* ``{{name}}.ngc``: synthesized RTL.
		* ``{{name}}.bld``: NGDBuild log.
		* ``{{name}}.ngd``: design database.
		* ``{{name}}_map.map``: MAP log.
		* ``{{name}}_map.mrp``: mapping report.
		* ``{{name}}_map.ncd``: mapped netlist.
		* ``{{name}}.pcf``: physical constraints.
		* ``{{name}}_par.par``: PAR log.
		* ``{{name}}_par_pad.txt``: I/O usage report.
		* ``{{name}}_par.ncd``: place and routed netlist.
		* ``{{name}}.drc``: DRC report.
		* ``{{name}}.bgn``: BitGen log.
		* ``{{name}}.bit``: binary bitstream with metadata.
		* ``{{name}}.bin``: raw binary bitstream.

	.. rubric:: Symbiflow toolchain

	Required tools:
		* ``symbiflow_synth``
		* ``symbiflow_pack``
		* ``symbiflow_place``
		* ``symbiflow_route``
		* ``symbiflow_write_fasm``
		* ``symbiflow_write_bitstream``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_SYMBIFLOW``, if present.

	Available overrides:
		* ``add_constraints``: inserts commands in XDC file.

	.. rubric:: yosys_nextpnr toolchain

	Required tools:
		* ``yosys``
		* ``nextpnr-xilinx``
		* ``fasm2frames``
		* ``xc7frames2bit``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_YOSYS_NEXTPNR``, if present.

	Available overrides:
		* ``nextpnr_db_dir``: overrides nextpnr-xilinx chip-db path.
		* ``prjxray_db_dir``: overrides prjxray db path prefix.
		* ``read_verilog_opts``: adds options for ``read_verilog`` Yosys command.
		* ``script_after_read``: inserts commands after ``read_verilog`` in Yosys script.
		* ``synth_opts``: adds options for ``synth_xilinx`` Yosys command.
		* ``script_after_synth``: inserts commands after ``synth_xilinx`` in Yosys script.
		* ``yosys_opts``: adds extra options for ``yosys``.
		* ``nextpnr_opts``: adds extra options for ``nextpnr-xilinx``.
		* ``fasm2frames_opts``: adds extra options for ``fasm2frames``.
		* ``xc7frames2bit_opts``: adds extra options for ``xc7frames2bit``.

	''' # noqa: E101

	toolchain = None # selected when creating platform

	@property
	@abstractmethod
	def device(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def package(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def speed(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	def _part(self) -> str:
		if self.family in { 'ultrascale', 'ultrascaleplus' }:
			return f'{self.device}-{self.package}-{self.speed}'
		else:
			return f'{self.device}{self.package}-{self.speed}'

	# Vivado templates

	_vivado_required_tools = [ 'vivado' ]
	_vivado_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			if [ -z "$BASH" ] ; then exec /bin/bash "$0" "$@"; fi
			[ -n "${{platform._toolchain_env_var}}" ] && . "${{platform._toolchain_env_var}}"
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.tcl': r'''
			# {{autogenerated}}
			create_project -force -name {{name}} -part {{platform._part}}
			{% for file in platform.iter_files(".v", ".sv", ".vhd", ".vhdl") -%}
				add_files {{file|tcl_escape}}
			{% endfor %}
			add_files {{name}}.v
			read_xdc {{name}}.xdc
			{% for file in platform.iter_files(".xdc") -%}
				read_xdc {{file|tcl_escape}}
			{% endfor %}
			{{get_override("script_after_read")|default("# (script_after_read placeholder)")}}
			synth_design -top {{name}} {{get_override("synth_design_opts")}}
			foreach cell [get_cells -quiet -hier -filter {torii.vivado.false_path == "TRUE"}] {
				set_false_path -to $cell
			}
			foreach cell [get_cells -quiet -hier -filter {torii.vivado.max_delay != ""}] {
				set clock [get_clocks -of_objects \
					[all_fanin -flat -startpoints_only [get_pin $cell/D]]]
				if {[llength $clock] != 0} {
					set_max_delay -datapath_only -from $clock \
						-to [get_cells $cell] [get_property torii.vivado.max_delay $cell]
				}
			}
			{{get_override("script_after_synth")|default("# (script_after_synth placeholder)")}}
			report_timing_summary -file {{name}}_timing_synth.rpt
			report_utilization -hierarchical -file {{name}}_utilization_hierarchical_synth.rpt
			report_utilization -file {{name}}_utilization_synth.rpt
			opt_design
			place_design
			{{get_override("script_after_place")|default("# (script_after_place placeholder)")}}
			report_utilization -hierarchical -file {{name}}_utilization_hierarchical_place.rpt
			report_utilization -file {{name}}_utilization_place.rpt
			report_io -file {{name}}_io.rpt
			report_control_sets -verbose -file {{name}}_control_sets.rpt
			report_clock_utilization -file {{name}}_clock_utilization.rpt
			route_design
			{{get_override("script_after_route")|default("# (script_after_route placeholder)")}}
			phys_opt_design
			report_timing_summary -no_header -no_detailed_paths
			write_checkpoint -force {{name}}_route.dcp
			report_route_status -file {{name}}_route_status.rpt
			report_drc -file {{name}}_drc.rpt
			report_methodology -file {{name}}_methodology.rpt
			report_timing_summary -datasheet -max_paths 10 -file {{name}}_timing.rpt
			report_power -file {{name}}_power.rpt
			{{get_override("script_before_bitstream")|default("# (script_before_bitstream placeholder)")}}
			write_bitstream -force -bin_file {{name}}.bit
			{{get_override("script_after_bitstream")|default("# (script_after_bitstream placeholder)")}}
			quit
		''',
		'{{name}}.xdc': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_property LOC {{pin_name}} [get_ports {{port_name|tcl_escape}}]
				{% for attr_name, attr_value in attrs.items() -%}
					set_property {{attr_name}} {{attr_value|tcl_escape}} [get_ports {{port_name|tcl_escape}}]
				{% endfor %}
			{% endfor %}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					create_clock -name {{port_signal.name|ascii_escape}} -period {{1000000000/frequency}} [get_ports {{port_signal.name|tcl_escape}}]
				{% else -%}
					create_clock -name {{net_signal.name|ascii_escape}} -period {{1000000000/frequency}} [get_nets {{net_signal|hierarchy("/")|tcl_escape}}]
				{% endif %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''' # noqa: E501
	}
	_vivado_command_templates = [
		r'''
		{{invoke_tool("vivado")}}
			{{verbose("-verbose")}}
			{{get_override("vivado_opts")|options}}
			-mode batch
			-log {{name}}.log
			-source {{name}}.tcl
		'''
	]

	# ISE toolchain

	_ise_required_tools = [
		'xst',
		'ngdbuild',
		'map',
		'par',
		'bitgen',
	]
	_ise_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			if [ -z "$BASH" ] ; then exec /bin/bash "$0" "$@"; fi
			[ -n "${{platform._toolchain_env_var}}" ] && . "${{platform._toolchain_env_var}}"
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.prj': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".vhd", ".vhdl") -%}
				vhdl work {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".v") -%}
				verilog work {{file}}
			{% endfor %}
			verilog work {{name}}.v
		''',
		'{{name}}.xst': r'''
			# {{autogenerated}}
			run
			-ifn {{name}}.prj
			-ofn {{name}}.ngc
			-top {{name}}
			-use_new_parser yes
			-p {{platform.device}}{{platform.package}}-{{platform.speed}}
			{{get_override("script_after_run")|default("# (script_after_run placeholder)")}}
		''',
		'{{name}}.ucf': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				{% set port_name = port_name|replace("[", "<")|replace("]", ">") -%}
				NET '{{port_name}}" LOC = {{pin_name}};
				{% for attr_name, attr_value in attrs.items() -%}
					NET '{{port_name}}" {{attr_name}} = {{attr_value}};
				{% endfor %}
			{% endfor %}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				NET '{{net_signal|hierarchy("/")}}" TNM_NET = "PRD{{net_signal|hierarchy("/")}}";
				TIMESPEC "TS{{net_signal|hierarchy("__")}}" = PERIOD "PRD{{net_signal|hierarchy("/")}}" {{1000000000/frequency}} ns HIGH 50%;
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''' # noqa: E501
	}
	_ise_command_templates = [
		r'''
		{{invoke_tool("xst")}}
			{{get_override("xst_opts")|options}}
			-ifn {{name}}.xst
		''',
		r'''
		{{invoke_tool("ngdbuild")}}
			{{quiet("-quiet")}}
			{{verbose("-verbose")}}
			{{get_override("ngdbuild_opts")|options}}
			-uc {{name}}.ucf
			{{name}}.ngc
		''',
		r'''
		{{invoke_tool("map")}}
			{{verbose("-detail")}}
			{{get_override("map_opts")|default([])|options}}
			-w
			-o {{name}}_map.ncd
			{{name}}.ngd
			{{name}}.pcf
		''',
		r'''
		{{invoke_tool("par")}}
			{{get_override("par_opts")|default([])|options}}
			-w
			{{name}}_map.ncd
			{{name}}_par.ncd
			{{name}}.pcf
		''',
		r'''
		{{invoke_tool("bitgen")}}
			{{get_override("bitgen_opts")|default(["-g Compress"])|options}}
			-w
			-g Binary:Yes
			{{name}}_par.ncd
			{{name}}.bit
		'''
	]

	# Symbiflow templates

	_symbiflow_part_map = {
		'xc7a35ticsg324-1L': 'xc7a35tcsg324-1', # Arty-A7
	}

	_symbiflow_required_tools = [
		'symbiflow_synth',
		'symbiflow_pack',
		'symbiflow_place',
		'symbiflow_route',
		'symbiflow_write_fasm',
		'symbiflow_write_bitstream',
	]
	_symbiflow_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.pcf': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_io {{port_name}} {{pin_name}}
			{% endfor %}
		''',
		'{{name}}.xdc': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				{% for attr_name, attr_value in attrs.items() -%}
					set_property {{attr_name}} {{attr_value}} [get_ports {{port_name|tcl_escape}} }]
				{% endfor %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''',
		'{{name}}.sdc': r'''
			# {{autogenerated}}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is none -%}
					create_clock -period {{1000000000/frequency}} {{net_signal.name|ascii_escape}}
				{% endif %}
			{% endfor %}
		'''
	}
	_symbiflow_command_templates = [
		r'''
		{{invoke_tool("symbiflow_synth")}}
			-t {{name}}
			-v {% for file in platform.iter_files(".v", ".sv", ".vhd", ".vhdl") -%} {{file}} {% endfor %} {{name}}.v
			-p {{platform._symbiflow_part_map.get(platform._part, platform._part)}}
			-x {{name}}.xdc
		''',
		r'''
		{{invoke_tool("symbiflow_pack")}}
			-e {{name}}.eblif
			-P {{platform._symbiflow_part_map.get(platform._part, platform._part)}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_place")}}
			-e {{name}}.eblif
			-p {{name}}.pcf
			-n {{name}}.net
			-P {{platform._symbiflow_part_map.get(platform._part, platform._part)}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_route")}}
			-e {{name}}.eblif
			-P {{platform._symbiflow_part_map.get(platform._part, platform._part)}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_write_fasm")}}
			-e {{name}}.eblif
			-P {{platform._symbiflow_part_map.get(platform._part, platform._part)}}
		''',
		r'''
		{{invoke_tool("symbiflow_write_bitstream")}}
			-f {{name}}.fasm
			-p {{platform._symbiflow_part_map.get(platform._part, platform._part)}}
			-b {{name}}.bit
		'''
	]

	# Yosys NextPNR prjxray templates

	_yosys_nextpnr_part_map = {
		'xc7a35ticsg324-1L':  'xc7a35tcsg324-1',  # Arty-A7 35t
		'xc7a100ticsg324-1L': 'xc7a100tcsg324-1', # Arty-A7 100t
	}

	_yosys_nextpnr_device = {
		'xc7a35ti':  'xc7a35t',
		'xc7a100ti': 'xc7a100t',
	}

	_yosys_nextpnr_family = {
		'xc7a': 'artix7',
		'xc7z': 'zynq7',
	}

	_yosys_nextpnr_required_tools = [
		'yosys',
		'nextpnr-xilinx',
		'fasm2frames',
		'xc7frames2bit'
	]

	_yosys_nextpnr_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			if [ -z "$BASH" ] ; then exec /bin/bash "$0" "$@"; fi
			[ -n "${{platform._toolchain_env_var}}" ] && . "${{platform._toolchain_env_var}}"
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.xdc': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				{% for attr_name, attr_value in attrs.items() -%}
					set_property {{attr_name}} {{attr_value}} [get_ports {{port_name|tcl_escape}}]
					set_property LOC {{pin_name}} [get_ports {{port_name|tcl_escape}}]
				{% endfor %}
			{% endfor %}
		''',
		'{{name}}.ys': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v") -%}
				read_verilog {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".sv") -%}
				read_verilog -sv {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".il") -%}
				read_rtlil {{file}}
			{% endfor %}
			read_verilog {{name}}.v
			{{get_override("script_after_read")|default("# (script_after_read placeholder)")}}
			synth_xilinx {{get_override("synth_opts")|options}} -flatten -abc9 -nobram -arch xc7 -top {{name}};
			{{get_override("script_after_synth")|default("# (script_after_synth placeholder)")}}
			write_json {{name}}.json
		'''
	}

	_yosys_nextpnr_command_templates = [
		r'''
		PRJXRAY_DB_DIR={{get_override("prjxray_db_dir")|default("/usr/share/nextpnr/prjxray-db")}}
		''',
		r'''
		NEXPNR_CHIPDB_DIR={{get_override("nextpnr_db_dir")|default("/usr/share/nextpnr/xilinx-chipdb")}}
		''',
		r'''
		PART={{platform._yosys_nextpnr_part_map.get(platform._part, platform._part)}}
		''',
		r'''
		FAMILY={{platform._yosys_nextpnr_family[platform.device[:4]]}}
		''',
		r'''
		{{invoke_tool("yosys")}}
			{{quiet("-q")}}
			{{get_override("yosys_opts")|options}}
			-l {{name}}.rpt
			{{name}}.ys
		''',
		r'''
		{{invoke_tool("nextpnr-xilinx")}}
			{{quiet("--quiet")}}
			{{get_override("nextpnr_xilinx_opts")|options}}
			--log {{name}}.tim
			--chipdb "$NEXPNR_CHIPDB_DIR/$PART.bin"
			--xdc {{name}}.xdc
			--json {{name}}.json
			--write {{name}}_routed.json
			--fasm {{name}}.fasm
		''',
		r'''
		{{invoke_tool("fasm2frames")}}
			{{get_override("fasm2frames_opts")}}
			--part $PART
			--db-root "$PRJXRAY_DB_DIR/$FAMILY" {{name}}.fasm > {{name}}.frames
		''',
		r'''
		{{invoke_tool("xc7frames2bit")}}
			{{get_override("xc7frames2bit_opts")}}
			--part_file "$PRJXRAY_DB_DIR/$FAMILY/$PART/part.yaml"
			--part_name $PART
			--frm_file {{name}}.frames
			--output_file {{name}}.bit
		''',
	]

	# Common logic

	def __init__(self, *, toolchain: Literal['Vivado', 'ISE', 'Symbiflow'] = None) -> None:
		super().__init__()

		# Determine device family.
		device = self.device.lower()
		# Remove the prefix.
		if device.startswith('xc'):
			device = device[2:]
		elif device.startswith('xa'):
			device = device[2:]
		elif device.startswith('xqr'):
			device = device[3:]
		elif device.startswith('xq'):
			device = device[2:]
		else:
			raise ValueError(f'Device \'{self.device}\' is not recognized')
		# Do actual name matching.
		if device.startswith('2vp'):
			self.family = 'virtex2p'
		elif device.startswith('2v'):
			self.family = 'virtex2'
		elif device.startswith('3sd'):
			self.family = 'spartan3adsp'
		elif device.startswith('3s'):
			if device.endswith('a'):
				self.family = 'spartan3a'
			elif device.endswith('e'):
				self.family = 'spartan3e'
			else:
				self.family = 'spartan3'
		elif device.startswith('4v'):
			self.family = 'virtex4'
		elif device.startswith('5v'):
			self.family = 'virtex5'
		elif device.startswith('6v'):
			self.family = 'virtex6'
		elif device.startswith('6s'):
			self.family = 'spartan6'
		elif device.startswith('7'):
			self.family = 'series7'
		elif device.startswith(('vu', 'ku')):
			if device.endswith('p'):
				self.family = 'ultrascaleplus'
			else:
				self.family = 'ultrascale'
		elif device.startswith(('zu', 'u', 'k26', 'au')):
			self.family = 'ultrascaleplus'
		elif device.startswith(('v', '2s')):
			# Match last to avoid conflict with ultrascale.
			# Yes, Spartan 2 is the same thing as Virtex.
			if device.endswith('e'):
				self.family = 'virtexe'
			else:
				self.family = 'virtex'


		ISE_FAMILIES = {
				'virtex', 'virtexe',
				'virtex2', 'virtex2p',
				'spartan3', 'spartan3e', 'spartan3a', 'spartan3adsp',
				'virtex4',
				'virtex5',
				'virtex6',
				'spartan6',
		}
		if toolchain is None:
			if self.family in ISE_FAMILIES:
				toolchain = 'ISE'
			else:
				toolchain = 'Vivado'

		assert toolchain in ('Vivado', 'ISE', 'Symbiflow', 'yosys_nextpnr')
		if toolchain == 'Vivado':
			if self.family in ISE_FAMILIES:
				raise ValueError(f'Family \'{self.family}\' is not supported by the Vivado toolchain, please use ISE instead')
		elif toolchain == 'ISE':
			if self.family not in ISE_FAMILIES and self.family != "series7":
				raise ValueError(f'Family \'{self.family}\' is not supported by the ISE toolchain, please use Vivado instead')
		elif toolchain == "Symbiflow":
			if self.family != 'series7':
				raise ValueError(f'Family \'{self.family}\' is not supported by the Symbiflow toolchain')
		elif toolchain == 'yosys_nextpnr':
			if self.family != 'series7':
				raise ValueError(f'Family \'{self.family}\' is not supported by the yosys nextpnr toolchain')
		self.toolchain = toolchain

	@property
	def required_tools(self) -> list[str]:
		if self.toolchain == 'Vivado':
			return self._vivado_required_tools
		if self.toolchain == 'ISE':
			return self._ise_required_tools
		if self.toolchain == 'Symbiflow':
			return self._symbiflow_required_tools
		if self.toolchain == 'yosys_nextpnr':
			return self._yosys_nextpnr_required_tools
		assert False

	@property
	def file_templates(self) -> dict[str, str]:
		if self.toolchain == 'Vivado':
			return self._vivado_file_templates
		if self.toolchain == 'ISE':
			return self._ise_file_templates
		if self.toolchain == 'Symbiflow':
			return self._symbiflow_file_templates
		if self.toolchain == 'yosys_nextpnr':
			return self._yosys_nextpnr_file_templates
		assert False

	@property
	def command_templates(self) -> list[str]:
		if self.toolchain == 'Vivado':
			return self._vivado_command_templates
		if self.toolchain == 'ISE':
			return self._ise_command_templates
		if self.toolchain == 'Symbiflow':
			return self._symbiflow_command_templates
		if self.toolchain == 'yosys_nextpnr':
			return self._yosys_nextpnr_command_templates
		assert False

	def create_missing_domain(self, name: str) -> Module:
		# Xilinx devices have a global write enable (GWE) signal that asserted during configuration
		# and de-asserted once it ends. Because it is an asynchronous signal (GWE is driven by logic
		# synchronous to configuration clock, which is not used by most designs), even though it is
		# a low-skew global network, its de-assertion may violate a setup/hold constraint with
		# relation to a user clock. The recommended solution is to use a BUFGCE driven by the EOS
		# signal (if available). For details, see:
		#   * https://www.xilinx.com/support/answers/44174.html
		#   * https://www.xilinx.com/support/documentation/white_papers/wp272.pdf

		STARTUP_PRIMITIVE = {
				'spartan6': 'STARTUP_SPARTAN6',
				'virtex4': 'STARTUP_VIRTEX4',
				'virtex5': 'STARTUP_VIRTEX5',
				'virtex6': 'STARTUP_VIRTEX6',
				'series7': 'STARTUPE2',
				'ultrascale': 'STARTUPE3',
				'ultrascaleplus': 'STARTUPE3',
		}

		if self.family not in STARTUP_PRIMITIVE or self.toolchain in ('Symbiflow', 'yosys_nextpnr'):
			# Spartan 3 and before lacks a STARTUP primitive with EOS output; use a simple ResetSynchronizer
			# in that case, as is the default.
			# Symbiflow does not support the STARTUPE2 primitive.
			return super().create_missing_domain(name)

		if name == 'sync' and self.default_clk is not None:
			clk_i = self.request(self.default_clk).i
			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i

			m = Module()
			ready = Signal()
			m.submodules += Instance(STARTUP_PRIMITIVE[self.family], o_EOS = ready)
			m.domains += ClockDomain('sync', reset_less = self.default_rst is None)
			if self.toolchain != 'Vivado':
				m.submodules += Instance('BUFGCE', i_CE = ready, i_I = clk_i, o_O = ClockSignal('sync'))
			elif self.family == 'series7':
				# Actually use BUFGCTRL configured as BUFGCE, since using BUFGCE causes
				# sim/synth mismatches with Vivado 2019.2, and the suggested workaround
				# (SIM_DEVICE parameter) breaks Vivado 2017.4.
				m.submodules += Instance(
					'BUFGCTRL',
					p_SIM_DEVICE = '7SERIES',
					i_I0         = clk_i,
					i_S0         = Const(1, 1),
					i_CE0        = ready,
					i_IGNORE0    = Const(0, 1),
					i_I1         = Const(1, 1),
					i_S1         = Const(0, 1),
					i_CE1        = Const(0, 1),
					i_IGNORE1    = Const(1, 1),
					o_O          = ClockSignal('sync')
				)
			else:
				m.submodules += Instance(
					'BUFGCE',
					p_SIM_DEVICE = 'ULTRASCALE',
					i_CE = ready,
					i_I = clk_i,
					o_O = ClockSignal('sync')
				)
			if self.default_rst is not None:
				m.submodules.reset_sync = ResetSynchronizer(rst_i, domain = 'sync')
			return m

	def add_clock_constraint(self, clock: Signal, frequency: Union[int, float]) -> None:
		super().add_clock_constraint(clock, frequency)
		clock.attrs['keep'] = 'TRUE'

	def _get_xdr_buffer(
		self, m: Module , pin: Pin, iostd, *, i_invert: bool = False, o_invert: bool = False
	) -> tuple[Optional[Signal], Optional[Signal], Optional[Signal]]:
		XFDDR_FAMILIES = {
			'virtex2',
			'virtex2p',
			'spartan3',
		}
		XDDR2_FAMILIES = {
			'spartan3e',
			'spartan3a',
			'spartan3adsp',
			'spartan6',
		}
		XDDR_FAMILIES = {
			'virtex4',
			'virtex5',
			'virtex6',
			'series7',
		}
		XDDRE1_FAMILIES = {
			'ultrascale',
			'ultrascaleplus',
		}

		def get_iob_dff(clk: Signal, d: Signal, q: Signal) -> None:
			# SDR I/O is performed by packing a flip-flop into the pad IOB.
			for bit in range(len(q)):
				m.submodules += Instance(
					'FDCE',
					a_IOB = 'TRUE',
					i_C = clk,
					i_CE = Const(1),
					i_CLR = Const(0),
					i_D = d[bit],
					o_Q = q[bit]
				)

		def get_dff(clk: Signal, d: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'FDCE',
					i_C = clk,
					i_CE = Const(1),
					i_CLR = Const(0),
					i_D = d[bit],
					o_Q = q[bit]
				)

		def get_ifddr(clk: Signal, io: Signal, q0: Signal, q1: Signal) -> None:
			assert self.family in XFDDR_FAMILIES
			for bit in range(len(q0)):
				m.submodules += Instance(
					'IFDDRCPE',
					i_C0 = clk,
					i_C1 = ~clk,
					i_CE = Const(1),
					i_CLR = Const(0),
					i_PRE = Const(0),
					i_D = io[bit],
					o_Q0 = q0[bit],
					o_Q1 = q1[bit]
				)

		def get_iddr2(clk: Signal, d: Signal, q0: Signal, q1: Signal, alignment) -> None:
			assert self.family in XDDR2_FAMILIES
			for bit in range(len(q0)):
				m.submodules += Instance(
					'IDDR2',
					p_DDR_ALIGNMENT = alignment,
					p_SRTYPE = 'ASYNC',
					p_INIT_Q0 = Const(0, 1),
					p_INIT_Q1 = Const(0, 1),
					i_C0 = clk,
					i_C1 = ~clk,
					i_CE = Const(1),
					i_S = Const(0),
					i_R = Const(0),
					i_D = d[bit],
					o_Q0 = q0[bit],
					o_Q1 = q1[bit]
				)

		def get_iddr(clk: Signal, d: Signal, q1: Signal, q2: Signal) -> None:
			assert self.family in XDDR_FAMILIES or self.family in XDDRE1_FAMILIES
			for bit in range(len(q1)):
				if self.family in XDDR_FAMILIES:
					m.submodules += Instance(
						'IDDR',
						p_DDR_CLK_EDGE = 'SAME_EDGE_PIPELINED',
						p_SRTYPE = 'ASYNC',
						p_INIT_Q1 = Const(0, 1),
						p_INIT_Q2 = Const(0, 1),
						i_C = clk,
						i_CE = Const(1),
						i_S = Const(0),
						i_R = Const(0),
						i_D = d[bit],
						o_Q1 = q1[bit],
						o_Q2 = q2[bit]
					)
				else:
					m.submodules += Instance(
						'IDDRE1',
						p_DDR_CLK_EDGE = 'SAME_EDGE_PIPELINED',
						p_IS_C_INVERTED = Const(0, 1),
						p_IS_CB_INVERTED = Const(1, 1),
						i_C = clk,
						i_CB = clk,
						i_R = Const(0),
						i_D = d[bit],
						o_Q1 = q1[bit],
						o_Q2 = q2[bit]
					)

		def get_fddr(clk: Signal, d0: Signal, d1: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				if self.family in XFDDR_FAMILIES:
					m.submodules += Instance(
						'FDDRCPE',
						i_C0 = clk,
						i_C1 = ~clk,
						i_CE = Const(1),
						i_PRE = Const(0),
						i_CLR = Const(0),
						i_D0 = d0[bit],
						i_D1 = d1[bit],
						o_Q = q[bit]
					)
				else:
					m.submodules += Instance(
						'ODDR2',
						p_DDR_ALIGNMENT = 'NONE',
						p_SRTYPE = 'ASYNC',
						p_INIT = Const(0, 1),
						i_C0 = clk,
						i_C1 = ~clk,
						i_CE = Const(1),
						i_S = Const(0),
						i_R = Const(0),
						i_D0 = d0[bit],
						i_D1 = d1[bit],
						o_Q = q[bit]
					)

		def get_oddr(clk: Signal, d1: Signal, d2: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				if self.family in XDDR2_FAMILIES:
					m.submodules += Instance(
						'ODDR2',
						p_DDR_ALIGNMENT = 'C0',
						p_SRTYPE = 'ASYNC',
						p_INIT = Const(0, 1),
						i_C0 = clk,
						i_C1 = ~clk,
						i_CE = Const(1),
						i_S = Const(0),
						i_R = Const(0),
						i_D0 = d1[bit],
						i_D1 = d2[bit],
						o_Q = q[bit]
					)
				elif self.family in XDDR_FAMILIES:
					m.submodules += Instance(
						'ODDR',
						p_DDR_CLK_EDGE = 'SAME_EDGE',
						p_SRTYPE = 'ASYNC',
						p_INIT = Const(0, 1),
						i_C = clk,
						i_CE = Const(1),
						i_S = Const(0),
						i_R = Const(0),
						i_D1 = d1[bit],
						i_D2 = d2[bit],
						o_Q = q[bit]
					)
				elif self.family in XDDRE1_FAMILIES:
					m.submodules += Instance(
						'ODDRE1',
						p_SRVAL = Const(0, 1),
						i_C = clk,
						i_SR = Const(0),
						i_D1 = d1[bit],
						i_D2 = d2[bit],
						o_Q = q[bit]
					)

		def get_ineg(y: Signal, invert: bool) -> Signal:
			if invert:
				a = Signal.like(y, name_suffix = '_n')
				m.d.comb += y.eq(~a)
				return a
			else:
				return y

		def get_oneg(a: Signal, invert: bool) -> Signal:
			if invert:
				y = Signal.like(a, name_suffix = '_n')
				m.d.comb += y.eq(~a)
				return y
			else:
				return a

		if 'i' in pin.dir:
			if pin.xdr < 2:
				pin_i  = get_ineg(pin.i,  i_invert)
			elif pin.xdr == 2:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
		if 'o' in pin.dir:
			if pin.xdr < 2:
				pin_o  = get_oneg(pin.o,  o_invert)
			elif pin.xdr == 2:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)

		i = o = t = None
		if 'i' in pin.dir:
			i = Signal(pin.width, name = f'{pin.name}_xdr_i')
		if 'o' in pin.dir:
			o = Signal(pin.width, name = f'{pin.name}_xdr_o')
		if pin.dir in ('oe', 'io'):
			t = Signal(1,         name = f'{pin.name}_xdr_t')

		if pin.xdr == 0:
			if 'i' in pin.dir:
				i = pin_i
			if 'o' in pin.dir:
				o = pin_o
			if pin.dir in ('oe', 'io'):
				t = ~pin.oe
		elif pin.xdr == 1:
			if 'i' in pin.dir:
				get_iob_dff(pin.i_clk, i, pin_i)
			if 'o' in pin.dir:
				get_iob_dff(pin.o_clk, pin_o, o)
			if pin.dir in ('oe', 'io'):
				get_iob_dff(pin.o_clk, ~pin.oe, t)
		elif pin.xdr == 2:
			# On Spartan 3E/3A, the situation with DDR registers is messy: while the hardware
			# supports same-edge alignment, it does so by borrowing the resources of the other
			# pin in the differential pair (if any).  Since we cannot be sure if the other pin
			# is actually unused (or if the pin is even part of a differential pair in the first
			# place), we only use the hardware alignment feature in two cases:
			#
			# - differential inputs (since the other pin's input registers will be unused)
			# - true differential outputs (since they use only one pin's output registers,
			#   as opposed to pseudo-differential outputs that use both)
			TRUE_DIFF_S3EA = {
				'LVDS_33', 'LVDS_25',
				'MINI_LVDS_33', 'MINI_LVDS_25',
				'RSDS_33', 'RSDS_25',
				'PPDS_33', 'PPDS_25',
				'TMDS_33',
			}
			DIFF_S3EA = TRUE_DIFF_S3EA | {
				'DIFF_HSTL_I',
				'DIFF_HSTL_III',
				'DIFF_HSTL_I_18',
				'DIFF_HSTL_II_18',
				'DIFF_HSTL_III_18',
				'DIFF_SSTL3_I',
				'DIFF_SSTL3_II',
				'DIFF_SSTL2_I',
				'DIFF_SSTL2_II',
				'DIFF_SSTL18_I',
				'DIFF_SSTL18_II',
				'BLVDS_25',
			}
			if 'i' in pin.dir:
				if self.family in XFDDR_FAMILIES:
					# First-generation input DDR register: basically just two FFs with opposite
					# clocks. Add a register on both outputs, so that they enter fabric on
					# the same clock edge, adding one cycle of latency.
					i0_ff = Signal.like(pin_i0, name_suffix = '_ff')
					i1_ff = Signal.like(pin_i1, name_suffix = '_ff')
					get_dff(pin.i_clk, i0_ff, pin_i0)
					get_dff(pin.i_clk, i1_ff, pin_i1)
					get_iob_dff(pin.i_clk, i, i0_ff)
					get_iob_dff(~pin.i_clk, i, i1_ff)
				elif self.family in XDDR2_FAMILIES:
					if self.family == 'spartan6' or iostd in DIFF_S3EA:
						# Second-generation input DDR register: hw realigns i1 to positive clock edge,
						# but also misaligns it with i0 input.  Re-register first input before it
						# enters fabric. This allows both inputs to enter fabric on the same clock
						# edge, and adds one cycle of latency.
						i0_ff = Signal.like(pin_i0, name_suffix = '_ff')
						get_dff(pin.i_clk, i0_ff, pin_i0)
						get_iddr2(pin.i_clk, i, i0_ff, pin_i1, 'C0')
					else:
						# No extra register available for hw alignment, use extra registers.
						i0_ff = Signal.like(pin_i0, name_suffix = '_ff')
						i1_ff = Signal.like(pin_i1, name_suffix = '_ff')
						get_dff(pin.i_clk, i0_ff, pin_i0)
						get_dff(pin.i_clk, i1_ff, pin_i1)
						get_iddr2(pin.i_clk, i, i0_ff, i1_ff, "NONE")
				else:
					# Third-generation input DDR register: does all of the above on its own.
					get_iddr(pin.i_clk, i, pin_i0, pin_i1)
			if "o" in pin.dir:
				if self.family in XFDDR_FAMILIES or self.family == 'spartan3e' or (
					self.family.startswith("spartan3a") and iostd not in TRUE_DIFF_S3EA
				):
					# For this generation, we need to realign o1 input ourselves.
					o1_ff = Signal.like(pin_o1, name_suffix = '_ff')
					get_dff(pin.o_clk, pin_o1, o1_ff)
					get_fddr(pin.o_clk, pin_o0, o1_ff, o)
				else:
					get_oddr(pin.o_clk, pin_o0, pin_o1, o)
			if pin.dir in ('oe', 'io'):
				if self.family == 'spartan6':
					get_oddr(pin.o_clk, ~pin.oe, ~pin.oe, t)
				else:
					get_iob_dff(pin.o_clk, ~pin.oe, t)
		else:
			assert False

		return (i, o, t)

	def _get_valid_xdrs(self) -> Union[
		tuple[Literal[0], Literal[1]],
		tuple[Literal[0], Literal[1], Literal[2]]
	]:
		if self.family in { 'virtex', 'virtexe' }:
			return (0, 1)
		else:
			return (0, 1, 2)

	def get_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD'), i_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IBUF',
				i_I = port.io[bit],
				o_O = i[bit]
			)
		return m

	def get_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended output', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD'), o_invert = invert)
		if self.toolchain != 'Symbiflow':
			for bit in range(pin.width):
				m.submodules[f'{pin.name}_{bit}'] = Instance(
					'OBUF',
					i_I = o[bit],
					o_O = port.io[bit]
				)
		else:
			m.d.comb += port.eq(self._invert_if(invert, o))
		return m

	def get_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		if self.toolchain == 'Symbiflow':
			return super().get_tristate(pin, port, attrs, invert)

		self._check_feature(
			'single-ended tristate', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD'), o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBUFT',
				i_T = t,
				i_I = o[bit],
				o_O = port.io[bit]
			)
		return m

	def get_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		if self.toolchain == 'Symbiflow':
			return super().get_input_output(pin, port, attrs, invert)

		self._check_feature(
			'single-ended input/output', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD'), i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IOBUF',
				i_T = t,
				i_I = o[bit],
				o_O = i[bit],
				io_IO = port.io[bit]
			)
		return m

	def get_diff_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		if self.toolchain == 'Symbiflow':
			return super().get_diff_input(pin, port, attrs, invert)

		self._check_feature(
			'differential input', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD', 'LVDS_25'), i_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IBUFDS',
				i_I = port.p[bit],
				i_IB = port.n[bit],
				o_O = i[bit]
			)
		return m

	def get_diff_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		if self.toolchain == 'Symbiflow':
			return super().get_diff_output(pin, port, attrs, invert)

		self._check_feature(
			'differential output', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD', 'LVDS_25'), o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBUFDS',
				i_I = o[bit],
				o_O = port.p[bit],
				o_OB = port.n[bit]
			)
		return m

	def get_diff_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		if self.toolchain == 'Symbiflow':
			return super().get_diff_tristate(pin, port, attrs, invert)

		self._check_feature(
			'differential tristate', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD', 'LVDS_25'), o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBUFTDS',
				i_T = t,
				i_I = o[bit],
				o_O = port.p[bit],
				o_OB = port.n[bit]
			)
		return m

	def get_diff_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		if self.toolchain == 'Symbiflow':
			return super().get_diff_input_output(pin, port, attrs, invert)

		self._check_feature(
			'differential input/output', pin, attrs, valid_xdrs = self._get_valid_xdrs(), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, attrs.get('IOSTANDARD', 'LVDS_25'), i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IOBUFDS',
				i_T = t,
				i_I = o[bit],
				o_O = i[bit],
				io_IO = port.p[bit],
				io_IOB = port.n[bit]
			)
		return m

	# The synchronizer implementations below apply two separate but related timing constraints.
	#
	# First, the ASYNC_REG attribute prevents inference of shift registers from synchronizer FFs,
	# and constraints the FFs to be placed as close as possible, ideally in one CLB. This attribute
	# only affects the synchronizer FFs themselves.
	#
	# Second, for Vivado only, the torii.vivado.false_path or torii.vivado.max_delay attribute
	# affects the path into the synchronizer. If maximum input delay is specified, a datapath-only
	# maximum delay constraint is applied, limiting routing delay (and therefore skew) at
	# the synchronizer input.  Otherwise, a false path constraint is used to omit the input path
	# from the timing analysis.

	def get_ff_sync(self, ff_sync) -> Module:
		m = Module()
		flops = [
			Signal(
				ff_sync.i.shape(), name = f'stage{index}',
				reset = ff_sync._reset, reset_less = ff_sync._reset_less,
				attrs = { 'ASYNC_REG': 'TRUE' }
			)
			for index in range(ff_sync._stages)
		]
		if self.toolchain == 'Vivado':
			if ff_sync._max_input_delay is None:
				flops[0].attrs['torii.vivado.false_path'] = 'TRUE'
			else:
				flops[0].attrs['torii.vivado.max_delay'] = str(ff_sync._max_input_delay * 1e9)
		elif ff_sync._max_input_delay is not None:
			raise NotImplementedError(
				f'Platform \'{type(self).__name__}\' does not support constraining input delay for FFSynchronizer'
			)
		for i, o in zip((ff_sync.i, *flops), flops):
			m.d[ff_sync._o_domain] += o.eq(i)
		m.d.comb += ff_sync.o.eq(flops[-1])
		return m


	def get_async_ff_sync(self, async_ff_sync) -> Module:
		m = Module()
		m.domains += ClockDomain('async_ff', async_reset = True, local = True)
		flops = [
			Signal(1,
				name = f'stage{index}', reset = 1,
				attrs = { 'ASYNC_REG': 'TRUE' }
			)
			for index in range(async_ff_sync._stages)
		]
		if self.toolchain == 'Vivado':
			if async_ff_sync._max_input_delay is None:
				flops[0].attrs['torii.vivado.false_path'] = 'TRUE'
			else:
				flops[0].attrs['torii.vivado.max_delay'] = str(async_ff_sync._max_input_delay * 1e9)
		elif async_ff_sync._max_input_delay is not None:
			raise NotImplementedError(
				f'Platform \'{type(self).__name__}\' does not support constraining input delay for AsyncFFSynchronizer'
			)

		for i, o in zip((0, *flops), flops):
			m.d.async_ff += o.eq(i)

		if async_ff_sync._edge == 'pos':
			m.d.comb += ResetSignal('async_ff').eq(async_ff_sync.i)
		else:
			m.d.comb += ResetSignal('async_ff').eq(~async_ff_sync.i)

		m.d.comb += [
			ClockSignal('async_ff').eq(ClockSignal(async_ff_sync._o_domain)),
			async_ff_sync.o.eq(flops[-1])
		]

		return m



# SPDX-License-Identifier: BSD-2-Clause

from re import Match, finditer, sub

__all__ = (
	'ascii_escape',
	'tcl_escape',
	'tcl_quote',
	'tool_env_var',
)

def ascii_escape(string: str) -> str:
	''' Apply escaping to turn any character that is not A-Za-z0-9_ into hex '''

	def esc_match(m: Match) -> str:
		if m.group(1) is not None:
			return f'_{ord(m.group(1)[0]):02x}_'
		return m.group(2)

	return ''.join(
		map(esc_match, finditer(r'([^A-Za-z0-9_])|(.)', string))
	)

def tcl_escape(string: str) -> str:
	''' Apply appropriate escaping for use in TCL scripts '''

	return '{' + sub(r'([{}\\])', r'\\\1', string) + '}'

def tcl_quote(string: str) -> str:
	''' Apply appropriate quoting for use in TCL scripts '''

	return '"' + sub(r'([$[\\"])', r'\\\1', string) + '"'

def tool_env_var(name: str) -> str:
	'''
	Convert the given tool name into a normalized version for
	environment variables.

	This generally means replacing any `-`'s with `_`'s and any
	`+`'s with `X`'s and converting the entire name to uppercase.

	Parameters
	----------
	name : str
		The tool/executable name.

	Returns
	-------
	str
		The name of the environment variable used to look for
		for tool overrides.

	'''

	return name.upper().replace('-', '_').replace('+', 'X')
